local modpath = ModPath 
local CHECK_COPS_INTERVAL = 0.25 --the frequency at which cop positions will be checked, in seconds

local NEARBY_RADIUS = 300 --the distance at which cops are considered "nearby", in centimeters

local NEARBY_ANGLE_FRONT = 45 --at angles below this number, cops are considered in front of you
local NEARBY_ANGLE_REAR = 135 --at angles greater than this number, cops are considered behind you
--to be considered next to you, cops must be between the two above angles
--(left vs right are determined automatically)

local MIN_REMINDER_TIME = 4 --enemy must remain on the same side and within the nearby radius for at least this many seconds in order to trigger the first "Still there" reminder
local REMINDER_INTERVAL = 4 --every (this number of seconds), nearby reminder will play, provided the enemy is still within the radius and on the same side

local CLEAR_BEFORE_LEFT = true
local CLEAR_BEFORE_RIGHT = true
local BEFORE_THREEWIDE = false
local CLEAR_LEFT = 0 --0 = clear
local CLEAR_RIGHT = 0

--minor perf improvement by making local references to globals
local mvec3_angle = mvector3.angle
local math_abs = math.abs
local math_sign = math.sign
local mvec3_dis = mvector3.distance





_G.RacingSpotter = {}

RacingSpotter.nearby_enemies = {
	--[[ex:
	[unitkey] = {
		unit = Unit, --enemy unit
		start_t = 123.456, --time that enemy started being nearby (leaving this radius removes the enemy from this table)
		side_start_t = 123.456,--time that enemy started being on this side; switching sides (passing you or falling behind you, etc) will reset this time
		side = "left", --current side that the enemy is on
		last_reminder_t = 123.456 --time that the player was last reminded of the enemy's continued proximity
	}
	--]]
}

RacingSpotter.last_update_t = -CHECK_COPS_INTERVAL

--event functions; called in situations as the names suggest
--i wrote some example log output; replace it as you like

function RacingSpotter:Init()
	blt.xaudio.setup()
	self.buffers = {
        left = XAudio.Buffer:new(modpath .. "sounds/left.ogg"),
        right = XAudio.Buffer:new(modpath .. "sounds/right.ogg"),
        wide = XAudio.Buffer:new(modpath .. "sounds/3wide.ogg"),
        clear = XAudio.Buffer:new(modpath .. "sounds/clear.ogg"),
		otherclear = XAudio.Buffer:new(modpath .. "sounds/otherclear.ogg"),
        holdline = XAudio.Buffer:new(modpath .. "sounds/holdline.ogg"),
        stillthere = XAudio.Buffer:new(modpath .. "sounds/stillthere.ogg")
    }
	self.soundsource = XAudio.Source:new()

	managers.hud:add_updator("racingspotter_updater",callback(RacingSpotter,RacingSpotter,"Update"))
	managers.hud:add_updator("racingspotter_audio_updater",callback(RacingSpotter,RacingSpotter,"UpdateSounds"))
end

function RacingSpotter:OnEnemyEnteredNearby(t,data)
	if data.side == "left" then 
		CLEAR_LEFT = CLEAR_LEFT + 1
		RacingSpotter:PlaySound("left")
	elseif data.side == "right" then 
		CLEAR_RIGHT = CLEAR_RIGHT + 1
		RacingSpotter:PlaySound("right")
	end
	--RacingSpotter:CheckStatus()
end

function RacingSpotter:OnEnemySwitchedSide(t,data,prev_side)
	blt.xaudio.setup()
	if data.side == "left" then 
		CLEAR_LEFT = CLEAR_LEFT + 1
		RacingSpotter:PlaySound("left")
	elseif data.side == "right" then
		CLEAR_RIGHT = CLEAR_RIGHT + 1
		RacingSpotter:PlaySound("right")
	elseif data.side == "rear" then 
		if prev_side == "left" then 
			CLEAR_LEFT = CLEAR_LEFT - 1
			RacingSpotter:PlaySound("clear")
		elseif prev_side == "right" then
			CLEAR_RIGHT = CLEAR_RIGHT - 1
			RacingSpotter:PlaySound("clear")
		end
	elseif data.side == "front" then
		if prev_side == "left" then 
			CLEAR_LEFT = CLEAR_LEFT - 1
			RacingSpotter:PlaySound("otherclear")
		elseif prev_side == "right" then
			CLEAR_RIGHT = CLEAR_RIGHT - 1
			RacingSpotter:PlaySound("otherclear")
		end
	end
	--RacingSpotter:CheckStatus()
	--log("Switched sides, now they're " .. tostring(data.side))
end

function RacingSpotter:OnEnemyLeftNearby(t,data)
	if data.side == "rear" then
		log("enemy moved behind")
	elseif data.side == "front" then 
		log("enemy moved in front")
	end
	if data.side == "left" then 
		CLEAR_LEFT = CLEAR_LEFT - 1
		RacingSpotter:PlaySound("clear")
	elseif data.side == "right" then
		CLEAR_RIGHT = CLEAR_RIGHT - 1
		RacingSpotter:PlaySound("clear")
	end
	--RacingSpotter:CheckStatus()
end

function RacingSpotter:RemindEnemyNearby(t,data)
	if t - data.side_start_t > 10 then 
		log("Still there. Been there a while now.")
		RacingSpotter:PlaySound("holdline")
	else
		log("Still there.")
		RacingSpotter:PlaySound("stillthere")
	end
end

--audio stuff
function RacingSpotter:CheckStatus() --Unused, math is broken (may fix at later date maybe but doubt it)
	local left_clear = true
	local right_clear = true
	local threewide = false

	if CLEAR_LEFT > 0 then left_clear = false end
	if CLEAR_RIGHT > 0 then right_clear = false end
	if (not left_clear) and (not right_clear) then threewide = true end

	--Threewide test
	if threewide ~= BEFORE_THREEWIDE then --New Situation
		if threewide then --are in threewide
			log("new threewide")
			RacingSpotter:PlaySound("wide")
			BEFORE_THREEWIDE = true
			CLEAR_BEFORE_LEFT = false
			CLEAR_BEFORE_RIGHT = false
		else
			log("just left a threewide")
			BEFORE_THREEWIDE = false
		end
	end

	--Left test
	if not threewide then
		if not (left_clear ~= CLEAR_BEFORE_LEFT) then --New Situation
			if left_clear == CLEAR_BEFORE_LEFT then
				log("Left now empty")
				RacingSpotter:PlaySound("clear")
				CLEAR_BEFORE_LEFT = true
			else
				log("New left enemy")
				RacingSpotter:PlaySound("left")
				CLEAR_BEFORE_LEFT = false
			end
		end
	end

	--Right test
	if left_clear and not threewide then
		if not (right_clear ~= CLEAR_BEFORE_RIGHT) then --New Situation
			if right_clear == CLEAR_BEFORE_RIGHT then
				log("Right now empty")
				RacingSpotter:PlaySound("clear")
				CLEAR_BEFORE_RIGHT = true
			else
				log("New right enemy")
				RacingSpotter:PlaySound("right")
				CLEAR_BEFORE_RIGHT = false
			end
		end
	end
	log ("Left enemy count = " .. tostring(CLEAR_LEFT) .. " Right enemy count = " .. tostring(CLEAR_RIGHT))
end

RacingSpotter.queued_sounds = {}
RacingSpotter.max_queue_size = 1
function RacingSpotter:PlaySound(snd_id)
    if not self.buffers[snd_id] then 
        log("No sound with id " .. tostring(snd_id))
        return
    end
    
    local num_queued = #self.queued_sounds
    if num_queued <= self.max_queue_size then 
        table.insert(self.queued_sounds,num_queued + 1,snd_id)
	else
		self.queued_sounds[num_queued] = snd_id
    end
end

function RacingSpotter:UpdateSounds(t,dt)
    local source = self.soundsource
    
    if source and not source._closed and source:get_state() ~= 1 then 
        for i,snd_id in ipairs(self.queued_sounds) do 
            local buffer = self.buffers[snd_id]
            if buffer then
                table.remove(self.queued_sounds,i)
                source:set_buffer(buffer)
                source:play()
                break
            else
                --log("ERROR: Buffer is not valid!")
            end
        end
    else
        --log("ERROR: Buffer is not valid!")
    end
end

--other stuff

function RacingSpotter:register_nearby_enemy(t,unit,side,distance)
	local unit_key = unit:key()
	local nearby_enemy_data = self:get_nearby_enemy(unit_key) 
	if nearby_enemy_data then 
		if nearby_enemy_data.side ~= side then 
			local prev_side = nearby_enemy_data.side
			nearby_enemy_data.side = side
			self:OnEnemySwitchedSide(t,nearby_enemy_data,prev_side)
			nearby_enemy_data.side_start_t = t
			nearby_enemy_data.last_reminder_t = t
		end
	else
		nearby_enemy_data = {
			unit = unit,
			start_t = t,
			side_start_t = t,
			side = side,
			last_reminder_t = t
		}
		self.nearby_enemies[unit_key] = nearby_enemy_data
		self:OnEnemyEnteredNearby(t,nearby_enemy_data)
	end
end

function RacingSpotter:unregister_nearby_enemy(t,unit,unit_key)
	if alive(unit) then 
		unit_key = unit_key or unit:key()
		local nearby_enemy_data = self:get_nearby_enemy(unit_key)
		self:OnEnemyLeftNearby(t,nearby_enemy_data)
		self.nearby_enemies[unit_key] = nil
	elseif unit_key then 
		self.nearby_enemies[unit_key] = nil
	end
end

function RacingSpotter:get_nearby_enemy(unit_key)
	return self.nearby_enemies[unit_key] 
end

function RacingSpotter.angle_from(a,b,c,d) -- converts to angle with ranges (-180 , 180); for result range 0-360, do +180 to result, or modulo 360
	a = a or "nil"
	b = b or "nil"
	c = c or "nil"
	d = d or "nil"
	local function do_angle(x1,y1,x2,y2)
		local angle = 0
		local x = x2 - x1 --x diff
		local y = y2 - y1 --y diff
		if x ~= 0 then 
			angle = math.atan(y / x) % 180
			if y == 0 then 
				if x > 0 then 
					angle = 180 --right
				else
					angle = 0 --left 
				end
			elseif y > 0 then 
				angle = angle - 180
			end
		else
			if y > 0 then
				angle = 270 --up
			else
				angle = 90 --down
			end
		end
		
		return angle
	end
	local vectype = type(Vector3())
	if (type(a) == vectype) and (type(b) == vectype) then  --vector pos diff
		return do_angle(a.x,a.y,b.x,b.y)
	elseif (type(a) == "number") and (type(b) == "number") and (type(c) == "number") and (type(d) == "number") then --manual x/y pos diff
		return do_angle(a,b,c,d)
	else
		return
	end
end

function RacingSpotter:Update(t,dt)
	self:UpdateSounds(t,dt)
	if t - RacingSpotter.last_update_t > CHECK_COPS_INTERVAL then
		RacingSpotter.last_update_t = t
		if managers.player then
			local player = managers.player:local_player()
			if alive(player) then 
				local player_pos = player:movement():m_pos()
				local nearby_enemies = World:find_units_quick("sphere", player_pos, NEARBY_RADIUS, managers.slot:get_mask("enemies"))
				
				for _,unit in pairs(nearby_enemies) do 
					local enemy_pos = unit:movement():m_pos()
					
					local player_facing_angle = player:camera():rotation():yaw()
					
					local angle_to_enemy = RacingSpotter.angle_from(player_pos,enemy_pos) - (player_facing_angle + 90)
					
					local angle_adjusted = (angle_to_enemy % 360) - 180
					
					local angle_sign = math_sign(angle_adjusted)
					local angle_abs = math_abs(angle_adjusted)
					
					if angle_abs < NEARBY_ANGLE_FRONT then 
						RacingSpotter:register_nearby_enemy(t,unit,"front")
						--self:unregister_nearby_enemy(t,unit,u_k)
					elseif angle_abs > NEARBY_ANGLE_REAR then 
						RacingSpotter:register_nearby_enemy(t,unit,"rear")
						--self:unregister_nearby_enemy(t,unit,u_k)
					else
						if angle_sign == -1 then 
							RacingSpotter:register_nearby_enemy(t,unit,"right")
						else
							RacingSpotter:register_nearby_enemy(t,unit,"left")
						end
					end
					
				end
				
				for u_k,data in pairs(self.nearby_enemies) do 
					local unit = data.unit
					if not alive(unit) then 
						self:unregister_nearby_enemy(t,unit,u_k)
					elseif mvec3_dis(unit:movement():m_pos(),player_pos) > NEARBY_RADIUS then 
						self:unregister_nearby_enemy(t,unit,u_k)
					else
						local side_start_t = data.side_start_t
						if t - data.side_start_t > MIN_REMINDER_TIME then 
							if data.last_reminder_t + REMINDER_INTERVAL < t then 
								data.last_reminder_t = t
								self:RemindEnemyNearby(t,data)
							end
						end
					end
				end
			end
		end
	
	end
end

Hooks:Add("BaseNetworkSessionOnLoadComplete","OnLoadComplete_RacingSpotter",function()
	RacingSpotter:Init()
end)